/**
 * Nehalem Uncore PAPI-C component
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, ZIH, TU Dresden, Federal Republic of Germany
 *
 * See the file COPYING in the package base directory for details
 **/

/** 
 * PAPI-C Component that enables access to Nehalem Uncore/Opteron NB Events via perfmon2 in systemwide mode through PAPI interface
 *  
 * Author: Daniel Molka (daniel.molka@tu-dresden.de)
 * Last upate: 01/20/10
 *
 * Description: 
 *  allows PAPI-C to access uncore events via perfmon2 in systemwide mode
 *  supports Intel Nehalem/Westmere Uncore events and AMD family 0fh/10h/11h NB events
 *  
 * Notes:
 *  - requires perfmon2 patched kernel, ENABLING THIS COMPONENT MAKES CPU COMPONENT UNUSEABLE !!!
 *  - requires cpuid and rdtsc instructions to be present
 *    - constant rate TSC required
 *    - disabling powermanagement and/or C-state might be required on certain CPUs (e.g. Athlon 64 X2, mobile Core 2 Duo)
 *  - requires sched_getcpu() (glibc >= 2.6)
 *  - sharing of eventsets between threads is possible, but private eventsets for each Thread are recommended
 *    in case of sharing eventsets:
 *      - setup Eventset and call PAPI_start() before parallel region
 *      - calls to PAPI_read() PAPI_reset() from multiple threads are allowed
 *      - do not call PAPI_stop() in one thread prior to another call to PAPI_read() from another thread
 *    in case of thread private eventsets:
 *      - it is allowed that threads call PAPI-stop() while others are still counting
 *      - it is recommended that no thread adds new events when other threads are already counting
 *  - multiple unit masks have to be specified in the correct order (as listed by pfmon -L)
 *  - thread migration is supported, however thread pinning is strongly recommended
 *  - best performance if only one thread per package (pinned to a certain core) uses PAPI
 */

/** TODO
 * - buffer overflow if code_to_name and code_to_descr write more than 128 byte into the output buffer, which is reported to be 1024 Byte
 * - error checking for all pfm_* calls
 * - better unit mask handling, allow arbitrary order, do not list all combinations in papi_native_avail
 * - reduce overhead for common case: speculative pfm_read(), error handling (get_cpu()... only in case of errors)
 * - do not migrate non-monitoring threads to counting core => signal monitoring thread to update counters
 */

/* eneble to print debug information (generates a lot of output)*/
//#define DEBUG

/*************************  INCLUDES SECTION  *********************************
*******************************************************************************/
/* needed for sched_{set|get}affinity() */
#ifndef _GNU_SOURCE
 #define _GNU_SOURCE
#endif

/* Headers required by PAPI */
#include "papi.h"
#include "papi_internal.h"
#include "papi_vector.h"
#include "papi_memory.h"

/* misc headers */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <utmpx.h>
#include <sched.h>

#ifdef DEBUG
#include <assert.h>
#endif

/* perfmon2 headers */
#include <perfmon/pfmlib.h>
#include <perfmon/pfmlib_intel_nhm.h>
#include <perfmon/pfmlib_amd64.h>
#include <perfmon/perfmon.h>

/*************************  DEFINES SECTION  ***********************************
*******************************************************************************/
/* this number assumes that there will never be more concurrent events than indicated */
#define UNCORE_MAX_COUNTERS 8
#define NEHALEM_UNCORE_MAX_COUNTERS 8
#define OPTERON_NB_MAX_COUNTERS 4

/* 
 * max age of counter values in cycles
 * each package has a monitoring thread that updates counters on every call of UNCORE_read()
 * other threads report stored values until the counter value is older than specified here
 * if TTL has expired other threads start updating the counters themselves (including thread migration to the counting core)
 * 
 * lower values result in higher accuracy but increase overhead as well
 */
#define _UNCORE_TTL 10000

/*
 * needed to choose between 32 and 64 bit assembly implementation for cpuid and rdtsc calls
 */
#if ((defined (__i386__))||(defined (__i386))||(defined (i386))||(defined (__i486__))||(defined (__i486))||(defined (i486))||(defined (__i586__))||(defined (__i586))||(defined (i586))||(defined (__i686__))||(defined (__i686))||(defined (i686)))
 #define _32_BIT
#else
  #if ((defined (__x86_64__))||(defined (__x86_64))||(defined (x86_64)))
    #define _64_BIT
  #else
    #define _arch_unknown
  #endif
#endif

/* supported architectures */
#define _TYPE_UNDEFINED 	0
#define _NEHALEM_UNCORE 	1
#define _OPTERON_NB 		2

/* Structure that stores private information of each event */
typedef struct UNCORE_register {
   /* This is used by the framework.It likes it to be !=0 to do somehting*/
   unsigned int selector;
} UNCORE_register_t;

/*
 * The following structures mimic the ones used by other components. It is more
 * convenient to use them like that as programming with PAPI makes specific
 * assumptions for them.
 */

/* This structure is used to build the table of events 
 * contains mapping from native event id to pfm event id 
 */
typedef struct UNCORE_native_event_entry {
   UNCORE_register_t resources;
   int event;
} UNCORE_native_event_entry_t;

/* copy and paste from ACPI example component, might be unnecessary */
typedef struct UNCORE_reg_alloc {
   UNCORE_register_t ra_bits;
} UNCORE_reg_alloc_t;

/* unused, all information stored in UNCORE_context_t */
typedef struct UNCORE_control_state {
} UNCORE_control_state_t;


typedef struct UNCORE_context {
  NativeInfo_t *native_events;                 /* selected events (max UNCORE_MAX_COUNTERS) */
  int *native_event_count;                     /* number of currently selected events in that context*/
  long long start_values[UNCORE_MAX_COUNTERS]; /* stores current values of all packages when starting or reseting counters */
  long long results[UNCORE_MAX_COUNTERS];
  int monitor; /* only one thread per package monitors uncore events others will read latest value from monitoring thread. */
  int package_id;                              /* package the thread currently runs on*/
  int core_id;                                 /* core the thread currently runs on*/
  int thread_id;                               /* thread id of the thread that the context belongs to*/
  int context_id;                              /* id of the context */
} UNCORE_context_t;

/* per package data needed by the component */
typedef struct UNCORE_package_data {
  volatile UNCORE_context_t *monitoring;                /* thread that currently monitors that package */
  volatile int loaded_core;				/* core the context was loaded on */
  volatile long long last_results[UNCORE_MAX_COUNTERS]; /* results read by other threads */
  volatile unsigned long long timestamp;                /* age of the last counter value */
  /* values needed by perfmon2 API */
  pfmlib_input_param_t inp; 
  void* mod_inp;
  pfmlib_output_param_t outp; 
  void* mod_outp;
  pfarg_ctx_t ctx; 
  pfarg_load_t load_arg; 
  pfarg_pmc_t *pc; 
  pfarg_pmd_t *pd; 
  int fd;
} UNCORE_package_data_t;

/*************************  GLOBALS SECTION  ***********************************
*******************************************************************************/
/* This table contains the Uncore/NB native events */
static UNCORE_native_event_entry_t *uncore_native_table;
/* number of events in the table*/
static unsigned NUM_EVENTS = 0;

/* required topology information */
 unsigned int num_packages;
 unsigned int num_cores;
UNCORE_package_data_t *package_data;     /* per package information */
unsigned int *package_id;                /* package ids of cores */
volatile unsigned long long _pfm_lock=0; /* lock variable */
/* available contexts - one for each thread */
 UNCORE_context_t **contexts;             
 int num_contexts=0;
/* available native event (events only, no unit_masks) */
 NativeInfo_t *native_events=NULL;
 int native_event_count=0;
 int max_counters; 

/* used to store Register content {R|E}AX, {R|E}BX, {R|E}CX and {R|E}DX 
 * needed by cpuid/rdtsc
 */
unsigned long long a,b,c,d;

#if defined _64_BIT
unsigned long long reg_a,reg_b,reg_c,reg_d;
#endif

#if defined _32_BIT
unsigned int reg_a,reg_b,reg_c,reg_d;
#endif

/** translate eventcode into human readable name */
int UNCORE_ntv_code_to_name(unsigned int EventCode, char *name, int len);

/*******************************************************************************
 ********  BEGIN FUNCTIONS  USED INTERNALLY SPECIFIC TO THIS COMPONENT *********
 ******************************************************************************/

#if defined _64_BIT
/** check if cpuid instruction is available */
static int has_cpuid()
{
  // all 64 Bit x86 CPUs support CPUID
  return 1;
}

/** call cpuid instruction */
static void cpuid(unsigned long long *a, unsigned long long *b, unsigned long long *c, unsigned long long *d)
{
  __asm__ __volatile__(
             "cpuid;"
           : "=a" (reg_a), "=b" (reg_b), "=c" (reg_c), "=d" (reg_d)
           : "a" (*a), "b" (*b), "c" (*c), "d" (*d)
     );
     *a=reg_a;
     *b=reg_b;
     *c=reg_c;
     *d=reg_d;
}

/** generate timestamp using rdtsc (see _UNCORE_TTL) */
static unsigned long long timestamp()
{
  __asm__ __volatile__("rdtsc;": "=a" (reg_a), "=d" (reg_d));
  return (reg_d<<32)|(reg_a&0xffffffffULL);
}

/** atomically replace lock variable with new value if it equals old value
 * @return returns old value on success, otherwise the current content of the lock variable is returned
 */
static unsigned long long cmpxchg(volatile unsigned long long *_pfm_lock,unsigned long long old,unsigned long long new)
{
    unsigned long long ret;
     __asm__ __volatile__(
             "lock cmpxchg %%rbx,(%%rcx);"
           : "=a" (ret)
           : "a" (old), "b" (new), "c" ((unsigned long long)_pfm_lock)
     );
     return ret;
}
#endif

#if defined _32_BIT
/** check if cpuid instruction is available */
static int has_cpuid()
{
   int flags_old,flags_new;

   __asm__ __volatile__(
           "pushfl;"
           "popl %%eax;"
           : "=a" (flags_old)
   );

   flags_new=flags_old;
   if (flags_old&(1<<21)) flags_new&=0xffdfffff; else flags_new|=(1<<21);

   __asm__ __volatile__(
           "pushl %%eax;"
           "popfl;"
           "pushfl;"
           "popl %%eax;"
           : "=a" (flags_new)
           : "a" (flags_new)
   );

   /* CPUID is supported if Bit 21 in the EFLAGS register can be changed */
   if (flags_new==flags_old) return 0; 
   else 
   {
     __asm__ __volatile__(
         "pushl %%eax;"
         "popfl;"
         :
         : "a" (flags_old)
     );
     return 1;
   }
}

/** call cpuid instruction */
static void cpuid(unsigned long long *a, unsigned long long *b, unsigned long long *c, unsigned long long *d)
{
  if (!has_cpuid()) return;
  else
  {

  __asm__ __volatile__(
             "cpuid;"
           : "=a" (reg_a), "=b" (reg_b), "=c" (reg_c), "=d" (reg_d)
           : "a" ((int)*a), "b" ((int)*b), "c" ((int)*c), "d" ((int)*d)
     );
     *a=(unsigned long long)reg_a;
     *b=(unsigned long long)reg_b;
     *c=(unsigned long long)reg_c;
     *d=(unsigned long long)reg_d;
  }
}

/** generate timestamp using rdtsc (see _UNCORE_TTL) */
static unsigned long long timestamp()
{
  __asm__ __volatile__("rdtsc;": "=a" (reg_a) , "=d" (reg_d));
  return (((unsigned long long)reg_d)<<32)+reg_a;
}

/** atomically replace lock variable with new value if it equals old value
 * @return returns old value on success, otherwise the current content of the lock variable is returned
 */
static unsigned long long cmpxchg(volatile unsigned long long *_pfm_lock,unsigned long long old,unsigned long long new)
{
    unsigned int ret;
     __asm__ __volatile__(
             "lock cmpxchg %%rbx,(%%rcx);"
           : "=a" (ret)
           : "a" ((unsigned int)old), "b" ((unsigned int)new), "c" ((unsigned int)_pfm_lock)
     );
     return (unsigned long long) ret;
}
#endif

/* prevent linker errors during compile, however component will not work (fail with error that cpuid is not supported) */
#if defined _arch_unknown
static int has_cpuid() {return 0;}
static void cpuid(unsigned long long *a, unsigned long long *b, unsigned long long *c, unsigned long long *d){return;}
static unsigned long long timestamp(){return 0;}
static unsigned long long cmpxchg(volatile unsigned long long *_pfm_lock,unsigned long long old,unsigned long long new){return 0;}
#endif

/** CPU identification using cpuid instruction if supported*/
static int get_cpu_vendor(char* vendor)
{
  if (!has_cpuid()) return -1;
  a=0;
  cpuid(&a,&b,&c,&d);
  *((unsigned int*)&(vendor[0]))=(int)b;
  *((unsigned int*)&(vendor[4]))=(int)d;
  *((unsigned int*)&(vendor[8]))=(int)c;
  vendor[12]='\0';

  return 0;
}
/** CPU identification using cpuid instruction if supported*/
static int get_cpu_family()
 {
      if (!has_cpuid()) return -1;
      a=0;
      cpuid(&a,&b,&c,&d);
      if (a>=1)
      {
        a=1;
        cpuid(&a,&b,&c,&d);

        return (((int)a>>8)&0xf)+(((int)a>>20)&0xff);
      }
      return -1;
 }
/** CPU identification using cpuid instruction if supported*/
static int get_cpu_model()
 {
      if (!has_cpuid()) return -1;
      a=0;
      cpuid(&a,&b,&c,&d);
      if (a>=1)
      {
        a=1;
        cpuid(&a,&b,&c,&d);

        return (((int)a>>4)&0xf)+(((int)a>>12)&0xf0);
      }
      return -1;
 }

/** determine number of CPUs in system */
static int num_cpus()
{
   int num;
   num=sysconf(_SC_NPROCESSORS_CONF);
   /*assume 1 if detection fails*/
   if (num<1) num=1;
   return num;
}

/** determine on which cpu the program is being run */
static int get_cpu()
{
  return sched_getcpu();
}

/** determine the physical package, a cpu belongs to */
static int get_pkg(int cpu)
{
  int pkg;
  int fd;
  char filename[256];
  char buffer[16];

  sprintf(filename,"/sys/devices/system/cpu/cpu%i/topology/physical_package_id",cpu);
  fd=open(filename,O_RDONLY);
  if (fd<=0) pkg=-1;
  else
  {
    if (read(fd,buffer,sizeof(buffer)/sizeof(char))>0) pkg=atoi(buffer);
    else pkg=-1;
    close(fd);
  }
  return pkg;
}

/* check if TSC is available */
static int has_rdtsc()
{
  if (!has_cpuid()) return 0;

  a=0;
  cpuid(&a,&b,&c,&d);
  if (a>=1)
  {
    a=1;
    cpuid(&a,&b,&c,&d);
    if ((int)d&(1<<4)) return 1;
  }

  return 0;

}

/* check if TSC is constant rate */
static int has_invariant_rdtsc()
{
   char tmp[12];
   int res=0;

   if ((has_rdtsc())&&(get_cpu_vendor((char*)&tmp[0])==0))
   {

     /* TSCs are usable if CPU supports only one frequency in C0 (no speedstep/Cool'n'Quite) 
        or if multiple frequencies are available and the constant/invariant TSC feature flag is set */
      
      if (!strcmp(&tmp[0],"GenuineIntel"))
      {
         /*check if Powermanagement and invariant TSC are supported*/
         if (has_cpuid())
         { 
           a=1;
           cpuid(&a,&b,&c,&d);
           /* no Frequency control */
           if ((!(d&(1<<22)))&&(!(c&(1<<7)))) res=1;
           a=0x80000000;
           cpuid(&a,&b,&c,&d);
           if (a >=0x80000007)
           {
              a=0x80000007;
              cpuid(&a,&b,&c,&d);
              /* invariant TSC */
              if (d&(1<<8)) res =1;
           }
         }
      }

      if (!strcmp(&tmp[0],"AuthenticAMD"))
      {
         /*check if Powermanagement and invariant TSC are supported*/
         if (has_cpuid())
         { 
           a=0x80000000;
           cpuid(&a,&b,&c,&d);
           if (a >=0x80000007)
           {
              a=0x80000007;
              cpuid(&a,&b,&c,&d);

              /* no Frequency control */
              if ((!(d&(1<<7)))&&(!(d&(1<<1)))) res=1;
              /* invariant TSC */
              if (d&(1<<8)) res =1;
           }
           /* assuming no frequency control if cpuid does not provide the extended function to test for it */
           else res=1;
         }
      }
   }
   return res;
}

/** called before entering critical sections, avoids interfering of threads on shared variables */
static void _uncore_lock(void)
{
  do
  {
    while(_pfm_lock!=0) usleep(1);
  }
  while (cmpxchg(&_pfm_lock,0,1)!=0);
  __asm__ __volatile__("mfence;");
}

/** release lock after critical section */
static void _uncore_unlock(void)
{
  __asm__ __volatile__("mfence;");
  _pfm_lock=0;
}

/*
 * finds correct context by thread id, creates new context if requiered
 *
 * if multiple threads share a global eventset (i.e UNCORE_init() is only called once) additional 
 * contexts are created for each thread, as component requires individual contexts
 */
static UNCORE_context_t* get_context_by_thread_id(UNCORE_context_t* ctx,int tid)
{
  int i;
  UNCORE_context_t* new;

  /* update thread_id of current context if it is uninitialized */
  if ((ctx->thread_id==PAPI_EMISC)||(ctx->thread_id==-1))
  {
    if (tid>=0) ctx->thread_id=tid;
    return ctx;
  }

  #ifdef DEBUG
  printf("replacing context for Thread-id %i with context of thread_id %i\n",ctx->thread_id,tid);
  #endif

  /* search existing contexts for the requested tid */
  for (i=0;i<num_contexts;i++) if (contexts[i]->thread_id==tid) return contexts[i];

  /* create new context if not found, copy information from original context */
  new=(UNCORE_context_t*)calloc(1,sizeof(UNCORE_context_t));
  memcpy(new,ctx,sizeof(UNCORE_context_t));
  num_contexts++;
  contexts=realloc(contexts,num_contexts*sizeof(UNCORE_context_t*));
  contexts[num_contexts-1]=new;
  new->context_id=num_contexts;
  new->thread_id=tid;
  new->monitor=0;

  return new;
}

/*
 * update counters of the package the thread associated to the context is running on
 * used by UNCORE_read()
 */
static int update_counters(volatile UNCORE_context_t *ctx)
{
  cpu_set_t mask,old_mask;
  int i,current_core_id;

  current_core_id=get_cpu();

  #ifdef DEBUG 
   printf("update counters on core %i in package %i\n",ctx->core_id,ctx->package_id);fflush(stdout);
   assert(_pfm_lock==1);
  #endif

  if (sched_getaffinity(0, sizeof(cpu_set_t),&old_mask)) return -1;
  CPU_ZERO(&mask);
  CPU_SET(ctx->core_id,&mask);
  if (sched_setaffinity(0, sizeof(cpu_set_t),&mask)) return -1;

  #ifdef DEBUG 
   printf("renew counters... ");fflush(stdout);
  #endif
  if (pfm_read_pmds(package_data[((UNCORE_context_t*)ctx)->package_id].fd, package_data[((UNCORE_context_t*)ctx)->package_id].pd, package_data[((UNCORE_context_t*)ctx)->package_id].outp.pfp_pmd_count) == -1) {
      printf("pfm_read_pmds failed\n");perror("");
      return -1;
  }
  for (i=0;i<package_data[((UNCORE_context_t*)ctx)->package_id].outp.pfp_pmd_count;i++) package_data[((UNCORE_context_t*)ctx)->package_id].last_results[i]=(long long)(package_data[((UNCORE_context_t*)ctx)->package_id].pd[i].reg_value);
  package_data[((UNCORE_context_t*)ctx)->package_id].timestamp=timestamp();

  if (current_core_id!=ctx->core_id)
  {
    CPU_ZERO(&mask);
    CPU_SET(current_core_id,&mask);
    sched_setaffinity(0, sizeof(cpu_set_t),&mask); /* migrate back to calling core */
  }
  sched_setaffinity(0, sizeof(cpu_set_t),&old_mask); /* restore affinity */
  return 0;
}

/*****************************************************************************
 *******************  BEGIN PAPI's COMPONENT REQUIRED FUNCTIONS  *************
 *****************************************************************************/

/****************************** Initialization *******************************/

/*
 * This is called whenever a thread is initialized, reserves memory for context
 * Note: threads are only initialized if they create their own eventset
 *       in case of multiple threads sharing one eventset this function is only called once
 *       (see get_context_by_thread_id() for the required workaround)
 */
int UNCORE_init(hwd_context_t *ctx)
{
   _uncore_lock(); 
   #ifdef DEBUG 
    printf("UNCORE_init %p ...",ctx);fflush(stdout);
   #endif

   memset(ctx,0,sizeof(UNCORE_context_t));
   ((UNCORE_context_t*)ctx)->context_id=num_contexts;
   ((UNCORE_context_t*)ctx)->native_event_count=malloc(sizeof(typeof(native_event_count)));
   *(((UNCORE_context_t*)ctx)->native_event_count)=0;
   num_contexts++;
   contexts=realloc(contexts,num_contexts*sizeof(UNCORE_context_t*));
   contexts[num_contexts-1]=(UNCORE_context_t*)ctx;
   ((UNCORE_context_t*)ctx)->thread_id=PAPI_thread_id();

   #ifdef DEBUG 
     printf(" OK\n");fflush(stdout);
   #endif
   _uncore_unlock();
   return PAPI_OK;
}

/* Initialize uncore counters, setup the function vector table
 * and get hardware information, this routine is called when the 
 * PAPI process is initialized (PAPI_library_init())
 */
int UNCORE_init_substrate()
{
  int i;                             /* loop index */
  int code;                          /* event code */
  int type=_TYPE_UNDEFINED;          /* processor type */
  unsigned int num_events=0;         /* number of native events*/
  char vendor[16];                   /* processor vendor */
  char event_name[PAPI_MAX_STR_LEN]; /* temp storage for eventnames */

  #ifdef DEBUG 
   printf("UNCORE_init_substrate...");fflush(stdout);
  #endif

  _uncore_lock();

  if (!has_rdtsc()){
    fprintf(stderr,"Error: rdtsc is not supported\n");fflush(stderr);
    _uncore_unlock();
    return PAPI_EBUG; /* no time stamp counter */
  }
  if (!has_invariant_rdtsc()) fprintf(stderr,"UNCORE_init_substarte - Warning: unreliable TSC\n");fflush(stderr);  

  /* check if processor is supported */
  get_cpu_vendor(vendor);
  if (!strcmp("GenuineIntel",vendor))
  {
    if (get_cpu_family()==6)
    {
      switch (get_cpu_model())
      {
        case 26: /* Bloomfield, Gainestown */
        case 30: /* Lynnfield, Clarksfield */
        case 37: /* Clarkdale, Arrandale */
        case 44: /* Gulftown, Westmere EP*/
        case 46: /* Beckton */
        /* add new models here */
           type=_NEHALEM_UNCORE;break;
        default: break;
      }
    }
  }
  if (!strcmp("AuthenticAMD",vendor))
  {
    switch (get_cpu_family())
    {
      case 15: /* K8 "Hammer" (family 0fh)*/
      case 16: /* K10 "Barcelona" (family 10h) */
      case 17: /* "Griffin" (family 11h) */
      /* add new families here */
         type=_OPTERON_NB;break;
      default: break;
    }
  }
  if (type==_TYPE_UNDEFINED) {
    if (has_cpuid()) fprintf(stderr,"Error: unsupported CPU - vendor: %s, Family: %i, Model: %i\n",vendor,get_cpu_family(),get_cpu_model());
    else fprintf(stderr,"Error: cpuid instruction not available\n");
    _uncore_unlock();
    return PAPI_EBUG; /* unsupported CPU */
  }

  #ifdef DEBUG
   if (type!=_TYPE_UNDEFINED)printf("\n  CPU vendor: %s - Family: %i, Model: %i",vendor,get_cpu_family(),get_cpu_model());
   if (type==_NEHALEM_UNCORE) printf("\n  using: \"Nehalem Uncore\" Method");
   if (type==_OPTERON_NB) printf("\n  using: \"Opteron Northbridge\" Method");
  #endif

   if (pfm_initialize() != PFMLIB_SUCCESS) {
     perror("Error: can not initialize libpfm\n");fflush(stderr);
     _uncore_unlock();
     return PAPI_EBUG; /* perfmon 2 internal error */
   }

   /* get topology information */
   num_cores=num_cpus();
   package_id=calloc(num_cores,sizeof(unsigned int));
   num_packages=0;
   for (i=0;i<num_cores;i++)
   {
     package_id[i]=get_pkg(i);
     if (package_id[i]>num_packages) num_packages=package_id[i];
   }
   num_packages++;

   /* reserve memory for the data stored per package */
   package_data=(UNCORE_package_data_t*)calloc(num_packages,sizeof(UNCORE_package_data_t));
   for (i=0;i<num_packages;i++) /* initialize contexts for all packages */
   {
     package_data[i].inp.pfp_dfl_plm     = PFM_PLM3|PFM_PLM0;      /* kernel and user mode */
     package_data[i].inp.pfp_flags       = PFMLIB_PFP_SYSTEMWIDE;  /* systemwide mode */
     package_data[i].ctx.ctx_flags       = PFM_FL_SYSTEM_WIDE;
     package_data[i].inp.pfp_event_count = 0;
     package_data[i].fd = pfm_create_context(&(package_data[i].ctx), NULL, NULL, 0);
     if (package_data[i].fd == -1) {
       fprintf(stderr,"Error: pfm_create_context failed\n");
       _uncore_unlock();
       return PAPI_EBUG; /* perfmon 2 internal error */
     }
     if (type==_NEHALEM_UNCORE)
     {
       package_data[i].mod_inp=calloc(1,sizeof(pfmlib_nhm_input_param_t));
       package_data[i].mod_outp=NULL;
     }
     if (type==_OPTERON_NB)
     {
       package_data[i].mod_inp=calloc(1,sizeof(pfmlib_amd64_input_param_t));
       package_data[i].mod_outp=calloc(1,sizeof(pfmlib_amd64_output_param_t));
     }
     package_data[i].loaded_core=-1;
   }

   #ifdef DEBUG 
    for(i=0;i<num_cores;i++) printf("\n  Uncore init substrate: Core %i running in package %i",i,package_id[i]);
    fflush(stdout);
   #endif

   /* Create events table */
   NUM_EVENTS=0;
   pfm_get_num_events(&num_events);
   /* allocate enough memory (only a fraction of the events are actually Uncore/NB events, however it's not know yet how many) */
   if ((uncore_native_table = (UNCORE_native_event_entry_t *)calloc(sizeof(UNCORE_native_event_entry_t),num_events)) == NULL) {
     perror("Could not get memory for events table");
     _uncore_unlock();
     return PAPI_EBUG;
   }

   /* 
    * iterate through list of events and select only Uncore/NB events
    * - Intel Nehalem/Westmere: clear naming convention allows easy desicion, all uncore events start with "UNC_" 
    * - AMD Family 0fh/10h/11h: filter events by list of NB events listed in AMD's BIOS and Kernel Developer Guide (BKDG)
    */
   if (type==_NEHALEM_UNCORE)
   {
    max_counters=NEHALEM_UNCORE_MAX_COUNTERS;
    for (i=0;i<num_events;i++)
    {
     //TODO replace with pfm_nhm_is_uncore()
     pfm_get_event_name(i,event_name,PAPI_MAX_STR_LEN-1);
     if (!strncmp(event_name,"UNC_",4))
     {
       uncore_native_table[NUM_EVENTS].event=i;
       uncore_native_table[NUM_EVENTS].resources.selector = i+1;
       NUM_EVENTS++;
      }
    }
   }
   if (type==_OPTERON_NB)
   {
     max_counters=OPTERON_NB_MAX_COUNTERS;
     for (i=0;i<num_events;i++)
     {
       pfm_get_event_code(i,&code);
       switch (code)
       {
          /* memory controller */
          case 0x00e0:
          case 0x00e1:
          case 0x00e2:
          case 0x00e3:
          case 0x00e4:
          case 0x00e8:
          case 0x00e9:
          case 0x00ea:
          case 0x00eb:
          case 0x00ec:
          case 0x00ee:
          case 0x01f0:
          /* Crossbar */
          case 0x01e0:
          case 0x01e1:
          case 0x01e2:
          case 0x01e3:
          case 0x01e4:
          case 0x01e5:
          case 0x01e6:
          case 0x01e7:
          /* HyperTransport */
          case 0x00f6:
          case 0x00f7:
          case 0x00f8:
          case 0x01f9:        
          /* L3 */
          case 0x04e0:
          case 0x04e1:
          case 0x04e2:
          case 0x04e3:
          case 0x04ed:
                        uncore_native_table[NUM_EVENTS].event=i;
                        uncore_native_table[NUM_EVENTS].resources.selector = i+1;
                        NUM_EVENTS++;
                        break;
          default:      break;
       }
     }
   }

   #ifdef DEBUG
    for (i=0;i<NUM_EVENTS;i++)
    {
      pfm_get_event_name(i,event_name,PAPI_MAX_STR_LEN-1);
      printf("\n  SYSWIDE_%s",event_name);fflush(stdout);
    }
    printf("\n  num_events: %i",num_events);
    printf("\n  Uncore/NB events: %i\n",NUM_EVENTS);
    printf(" OK\n");fflush(stdout);
   #endif

   /* reserve lowest Byte for unit mask */
   NUM_EVENTS=NUM_EVENTS<<8;

   /*
    * uncore_native_table only includes a list of Uncore/NB event codes and their corresponding pfm event id
    * unit masks are handled separately (lower byte in event)
    */

   _uncore_unlock();
   return PAPI_OK;
}

/* unused */
int UNCORE_init_control_state(hwd_control_state_t *ctrl){
   return PAPI_OK;
}

/******* Control of counters (Reading/Writing/Starting/Stopping/Setup) *******/

/*
 * start counting of Uncore/NB events (PAPI_start() was called in application)
 */
int UNCORE_start(hwd_context_t *ctx, hwd_control_state_t *ctrl){
   cpu_set_t mask,old_mask;
   int i;

   _uncore_lock();

   /* find correct context if a thread other than it's creator trys to use it*/
   if (((UNCORE_context_t*)ctx)->thread_id!=PAPI_thread_id()) ctx=get_context_by_thread_id(((UNCORE_context_t*)ctx),PAPI_thread_id());

   ((UNCORE_context_t*)ctx)->core_id=get_cpu();
   ((UNCORE_context_t*)ctx)->package_id=package_id[((UNCORE_context_t*)ctx)->core_id];
   if (sched_getaffinity(0, sizeof(cpu_set_t),&old_mask)) {_uncore_unlock();return PAPI_EBUG;}

   #ifdef DEBUG 
    printf("UNCORE_start...");fflush(stdout);
    assert(((UNCORE_context_t*)ctx)->core_id>=0);
    assert(((UNCORE_context_t*)ctx)->core_id<num_cpus());
    assert(((UNCORE_context_t*)ctx)->package_id>=0);
    assert(CPU_ISSET(((UNCORE_context_t*)ctx)->core_id,&old_mask));
   #endif

   /* 
    * bind to certain core for loading the perfmon context 
    * all further operations on that context have to be performed on the same core until unloading the context
    */
   CPU_ZERO(&mask);
   CPU_SET(((UNCORE_context_t*)ctx)->core_id,&mask);
   sched_setaffinity(0, sizeof(cpu_set_t),&mask);

   /* try to become the monitoring thread for that package if it does not already have one */
   if (package_data[((UNCORE_context_t*)ctx)->package_id].monitoring==NULL)
   {
     package_data[((UNCORE_context_t*)ctx)->package_id].load_arg.load_pid = ((UNCORE_context_t*)ctx)->core_id; /* set pid to core_id (systemwide mode) */
     if (pfm_load_context(package_data[((UNCORE_context_t*)ctx)->package_id].fd, &(package_data[((UNCORE_context_t*)ctx)->package_id].load_arg)) != PFMLIB_SUCCESS) {
       printf("Error: pfm_load_context failed\n");
       perror("");
       _uncore_unlock();
       return PAPI_EBUG; /* perfmon2 internal error */
     }
     else
     { 
       ((UNCORE_context_t*)ctx)->monitor=1;
       package_data[((UNCORE_context_t*)ctx)->package_id].monitoring=(UNCORE_context_t*)ctx;
       package_data[((UNCORE_context_t*)ctx)->package_id].loaded_core=((UNCORE_context_t*)ctx)->core_id;
       
       pfm_start(package_data[((UNCORE_context_t*)ctx)->package_id].fd, NULL);
     }
   } 
   else ((UNCORE_context_t*)ctx)->monitor=0;

   #ifdef DEBUG 
    assert(package_data[((UNCORE_context_t*)ctx)->package_id].monitoring!=0);
   #endif
   /* update counters and record start values */
   if (update_counters(package_data[((UNCORE_context_t*)ctx)->package_id].monitoring)!=0) {_uncore_unlock();return PAPI_EBUG;}
   for (i=0;i<UNCORE_MAX_COUNTERS;i++)((UNCORE_context_t*)ctx)->start_values[i]=package_data[((UNCORE_context_t*)ctx)->package_id].last_results[i];

   sched_setaffinity(0, sizeof(cpu_set_t),&old_mask); /* restore affinity */
   #ifdef DEBUG 
    printf(" OK\n");fflush(stdout);
   #endif

   _uncore_unlock();
   return PAPI_OK;
}

/*
 * stop counting (i.e. PAPI_stop() was called)
 */
int UNCORE_stop(hwd_context_t *ctx, hwd_control_state_t *ctrl)
{
   #ifdef DEBUG 
    printf("UNCORE_stop...");fflush(stdout);
   #endif

   _uncore_lock();

   /* find correct context if a thread other than it's creator trys to use it*/
   if (((UNCORE_context_t*)ctx)->thread_id!=PAPI_thread_id()) ctx=get_context_by_thread_id(((UNCORE_context_t*)ctx),PAPI_thread_id());

   /* update monitoring information */
   if (((UNCORE_context_t*)ctx)->monitor==1)
   {
     ((UNCORE_context_t*)ctx)->monitor=0;
     package_data[((UNCORE_context_t*)ctx)->package_id].monitoring=NULL;  
   }
   #ifdef DEBUG 
    printf(" OK\n");fflush(stdout);
   #endif
   _uncore_unlock();

   return PAPI_OK;
}

/*
 * read counters (i.e. PAPI_{read|stop}() was called)
 * TODO - try to speculatively read counters first (only works if thread was not migrated), error detection (getcpu() only in case of errors)
 */
int UNCORE_read(hwd_context_t *ctx, hwd_control_state_t *ctrl, long_long **events, int flags)
{
   cpu_set_t mask,old_mask;                  /* needed to change and restore affinity */
   int current_core_id,current_package_id;   /* current core (that we will pin to during read) and package_id (detect migration between packages) */
   int i,package_change=0;

   _uncore_lock();

   /* find correct context if a thread other than it's creator trys to use it*/
   if (((UNCORE_context_t*)ctx)->thread_id!=PAPI_thread_id()) ctx=get_context_by_thread_id(((UNCORE_context_t*)ctx),PAPI_thread_id());

   current_core_id=get_cpu();
   current_package_id=package_id[current_core_id];
   if (sched_getaffinity(0, sizeof(cpu_set_t),&old_mask)) {_uncore_unlock(); return PAPI_EBUG;}

   #ifdef DEBUG 
    printf("UNCORE_read... ");fflush(stdout);
    assert(CPU_ISSET(current_core_id,&old_mask));
   #endif

   /* 
    * if threads were migrated to a new core or even package some cleanup is required 
    * this includes high overhead and should be avoided 
    */

   /* intra-package migration of counting thread => unload context on old core and restart it on new core (perfmon2 allows only one context per package) */
   if ((current_core_id!=((UNCORE_context_t*)ctx)->core_id)&&(current_package_id==((UNCORE_context_t*)ctx)->package_id)&&(((UNCORE_context_t*)ctx)->monitor==1)) 
   {
     if (update_counters((UNCORE_context_t*)ctx)!=0) {_uncore_unlock();return PAPI_EBUG;}
     /*unload context on old CPU*/
     CPU_ZERO(&mask);
     CPU_SET(((UNCORE_context_t*)ctx)->core_id,&mask);
     sched_setaffinity(0, sizeof(cpu_set_t),&mask);
     pfm_stop(package_data[((UNCORE_context_t*)ctx)->package_id].fd);
     pfm_unload_context(package_data[((UNCORE_context_t*)ctx)->package_id].fd);
     /*load context on new CPU*/
     CPU_ZERO(&mask);
     CPU_SET(current_core_id,&mask);
     sched_setaffinity(0, sizeof(cpu_set_t),&mask);
     package_data[((UNCORE_context_t*)ctx)->package_id].load_arg.load_pid = current_core_id;
     if (pfm_load_context(package_data[((UNCORE_context_t*)ctx)->package_id].fd, &(package_data[((UNCORE_context_t*)ctx)->package_id].load_arg)) == -1) {
        printf("Warning - core change: pfm_load_context failed: fd=%i\n",package_data[((UNCORE_context_t*)ctx)->package_id].fd);
        ((UNCORE_context_t*)ctx)->monitor=0;
        package_data[current_package_id].monitoring=NULL;
        package_data[current_package_id].loaded_core=-1;
     }
     else
     { 
       pfm_start(package_data[((UNCORE_context_t*)ctx)->package_id].fd, NULL);
       package_data[current_package_id].loaded_core=current_core_id;
     }
   }
   /* inter-package migration of counting thread => leave context on old core/package loaded, not starting on new package (the case that there is no monitoring thread on new package is handled later)*/ 
   else if((current_package_id!=((UNCORE_context_t*)ctx)->package_id)&&(((UNCORE_context_t*)ctx)->monitor==1)) 
   {
     if (update_counters(package_data[((UNCORE_context_t*)ctx)->package_id].monitoring)!=0) {_uncore_unlock();return PAPI_EBUG;}
     for (i=0;i<UNCORE_MAX_COUNTERS;i++) ((UNCORE_context_t*)ctx)->start_values[i]=package_data[((UNCORE_context_t*)ctx)->package_id].last_results[i]-((UNCORE_context_t*)ctx)->start_values[i];
     package_change=1;

     if (update_counters((UNCORE_context_t*)ctx)!=0) {_uncore_unlock();return PAPI_EBUG;}
     package_data[((UNCORE_context_t*)ctx)->package_id].monitoring=NULL;
     package_data[((UNCORE_context_t*)ctx)->package_id].loaded_core=((UNCORE_context_t*)ctx)->core_id;
     if (((UNCORE_context_t*)ctx)->monitor>0) ((UNCORE_context_t*)ctx)->monitor--;

     CPU_ZERO(&mask);
     CPU_SET(current_core_id,&mask);
     sched_setaffinity(0, sizeof(cpu_set_t),&mask);
   }
   /* inter-package migration of non counting thread => recalculate start values as they depend on the current package */
   else if((current_package_id!=((UNCORE_context_t*)ctx)->package_id)&&(((UNCORE_context_t*)ctx)->monitor==0)) 
   {
     package_change=1;
     if (package_data[((UNCORE_context_t*)ctx)->package_id].monitoring!=0)
     {
      if (update_counters(package_data[((UNCORE_context_t*)ctx)->package_id].monitoring)!=0) {_uncore_unlock();return PAPI_EBUG;}
     }
     else /* no monitoring thread exists on old package (unlikely) */
     {
       /* check if a pfm context is still loaded on old package */
       if (package_data[((UNCORE_context_t*)ctx)->package_id].loaded_core==-1)
       {
          /* no loaded context => load context on old package */
          CPU_ZERO(&mask);
          CPU_SET(((UNCORE_context_t*)ctx)->core_id,&mask);
          sched_setaffinity(0, sizeof(cpu_set_t),&mask);

          package_data[((UNCORE_context_t*)ctx)->package_id].load_arg.load_pid = ((UNCORE_context_t*)ctx)->core_id;
          if (pfm_load_context(package_data[((UNCORE_context_t*)ctx)->package_id].fd, &(package_data[((UNCORE_context_t*)ctx)->package_id].load_arg)) == -1) {
               fprintf(stderr,"Error: core change - pfm_load_context failed: fd=%i\n",package_data[((UNCORE_context_t*)ctx)->package_id].fd);
               _uncore_unlock();return PAPI_EBUG;
          }
          pfm_start(package_data[((UNCORE_context_t*)ctx)->package_id].fd, NULL);
          package_data[((UNCORE_context_t*)ctx)->package_id].loaded_core=((UNCORE_context_t*)ctx)->core_id;
       }
       else
       {
          /* otherwise reuse loaded context on old package */
          CPU_ZERO(&mask);
          CPU_SET(package_data[((UNCORE_context_t*)ctx)->package_id].loaded_core,&mask);
          sched_setaffinity(0, sizeof(cpu_set_t),&mask);
          ((UNCORE_context_t*)ctx)->core_id=package_data[((UNCORE_context_t*)ctx)->package_id].loaded_core;
       }
       
       /* update counters */
       if (update_counters(((UNCORE_context_t*)ctx))!=0) {_uncore_unlock();return PAPI_EBUG;}

       CPU_ZERO(&mask);
       CPU_SET(current_core_id,&mask);
       sched_setaffinity(0, sizeof(cpu_set_t),&mask);       
     }
     for (i=0;i<UNCORE_MAX_COUNTERS;i++) ((UNCORE_context_t*)ctx)->start_values[i]=package_data[((UNCORE_context_t*)ctx)->package_id].last_results[i]-((UNCORE_context_t*)ctx)->start_values[i];

     CPU_ZERO(&mask);
     CPU_SET(current_core_id,&mask);
     sched_setaffinity(0, sizeof(cpu_set_t),&mask);
   }
   /* counting thread that did not migrate, or non-counting thread that stayed on the same package */
   else
   {
     CPU_ZERO(&mask);
     CPU_SET(current_core_id,&mask);
     sched_setaffinity(0, sizeof(cpu_set_t),&mask);
   }

   /* now thread is back on and pinned to the core it ran on when calling this function on after doing all necessary cleanup on the previous core */

   ((UNCORE_context_t*)ctx)->core_id=current_core_id;
   ((UNCORE_context_t*)ctx)->package_id=current_package_id;

   /*counting thread has been stoped or scheduled to another package, thus we need a new thread taking over the counting*/
   if (package_data[((UNCORE_context_t*)ctx)->package_id].monitoring==NULL) 
   {
     #ifdef DEBUG 
      printf("no monitoring thread\n");
     #endif
     /* check if a pfm context is still loaded on old package */
     if (package_data[((UNCORE_context_t*)ctx)->package_id].loaded_core==-1)
     {
        /* no loaded context => load context on old package */
        package_data[((UNCORE_context_t*)ctx)->package_id].load_arg.load_pid = ((UNCORE_context_t*)ctx)->core_id;
        if (pfm_load_context(package_data[((UNCORE_context_t*)ctx)->package_id].fd, &(package_data[((UNCORE_context_t*)ctx)->package_id].load_arg)) == -1) {
             fprintf(stderr,"Error: core change - pfm_load_context failed: fd=%i\n",package_data[((UNCORE_context_t*)ctx)->package_id].fd);
             _uncore_unlock();return PAPI_EBUG;
        }
        pfm_start(package_data[((UNCORE_context_t*)ctx)->package_id].fd, NULL);
        ((UNCORE_context_t*)ctx)->monitor=1;
        package_data[((UNCORE_context_t*)ctx)->package_id].monitoring=((UNCORE_context_t*)ctx);
        package_data[((UNCORE_context_t*)ctx)->package_id].loaded_core=((UNCORE_context_t*)ctx)->core_id;
     }
     else if (package_data[((UNCORE_context_t*)ctx)->package_id].loaded_core!=((UNCORE_context_t*)ctx)->core_id)
     {
        /* context loaded on other core => stop context on old core and restart on current core */
        CPU_ZERO(&mask);
        CPU_SET(package_data[((UNCORE_context_t*)ctx)->package_id].loaded_core,&mask);
        sched_setaffinity(0, sizeof(cpu_set_t),&mask);
        ((UNCORE_context_t*)ctx)->core_id=package_data[((UNCORE_context_t*)ctx)->package_id].loaded_core;
        if (update_counters((UNCORE_context_t*)ctx)!=0) {_uncore_unlock();return PAPI_EBUG;}
        /*unload context on old CPU*/
        pfm_stop(package_data[((UNCORE_context_t*)ctx)->package_id].fd);
        pfm_unload_context(package_data[((UNCORE_context_t*)ctx)->package_id].fd);
        /*load context on new CPU*/
        CPU_ZERO(&mask);
        CPU_SET(current_core_id,&mask);
        sched_setaffinity(0, sizeof(cpu_set_t),&mask);
        package_data[((UNCORE_context_t*)ctx)->package_id].load_arg.load_pid = current_core_id;
        if (pfm_load_context(package_data[((UNCORE_context_t*)ctx)->package_id].fd, &(package_data[((UNCORE_context_t*)ctx)->package_id].load_arg)) == -1) {
          printf("Warning - core change: pfm_load_context failed: fd=%i\n",package_data[((UNCORE_context_t*)ctx)->package_id].fd);
          ((UNCORE_context_t*)ctx)->monitor=0;
          package_data[current_package_id].monitoring=NULL;
          package_data[current_package_id].loaded_core=-1;
       }
       else
       { 
         pfm_start(package_data[((UNCORE_context_t*)ctx)->package_id].fd, NULL);
         package_data[current_package_id].loaded_core=current_core_id;
         package_data[current_package_id].monitoring=(UNCORE_context_t*)ctx;
         ((UNCORE_context_t*)ctx)->monitor=1;
         ((UNCORE_context_t*)ctx)->core_id=current_core_id;
       }
     }
     else
     {
       /* otherwise reuse loaded context on old package */
       ((UNCORE_context_t*)ctx)->monitor=1;
       package_data[current_package_id].monitoring=((UNCORE_context_t*)ctx);
     }
   }

   #ifdef DEBUG
    else printf("monitoring thread: %i\n",package_data[((UNCORE_context_t*)ctx)->package_id].monitoring->thread_id);
   #endif

   /* monitoring thread always reads counters */
   if (((UNCORE_context_t*)ctx)->monitor==1)
   {
    if (update_counters((UNCORE_context_t*)ctx)!=0) {_uncore_unlock();return PAPI_EBUG;}
   }

   #ifdef DEBUG 
    assert(package_data[((UNCORE_context_t*)ctx)->package_id].monitoring!=0);
   #endif

   /* non monitoring threads use last result until TTL is reached */
   if ((((UNCORE_context_t*)ctx)->monitor==0) && ((package_data[((UNCORE_context_t*)ctx)->package_id].timestamp<timestamp()+_UNCORE_TTL)||(package_change)))
   { /* values are to old, migrate to counting core and renew counters using other threads context - high overhead -> choose _UNCORE_TTL as high as possible*/
     //TODO do not migrate, signal monitoring thread to update counters
     if (update_counters(package_data[((UNCORE_context_t*)ctx)->package_id].monitoring)!=0) {_uncore_unlock();return PAPI_EBUG;}
   }

   /* recalculate start values if migrated to a new package */
   if (package_change)
   {
     for (i=0;i<UNCORE_MAX_COUNTERS;i++) ((UNCORE_context_t*)ctx)->start_values[i]=(long long)package_data[((UNCORE_context_t*)ctx)->package_id].last_results[i]-((UNCORE_context_t*)ctx)->start_values[i];
   }

   /* calculate results */
   //TODO check if threads with different eventsets get the wanted results (should be handled by selector ?)
   for (i=0;i<UNCORE_MAX_COUNTERS;i++) ((UNCORE_context_t*)ctx)->results[i] = package_data[((UNCORE_context_t*)ctx)->package_id].last_results[i] - ((UNCORE_context_t*)ctx)->start_values[i];
   *events=((UNCORE_context_t*)ctx)->results;

   #ifdef DEBUG 
    for (i=0;i<package_data[((UNCORE_context_t*)ctx)->package_id].outp.pfp_pmd_count;i++) printf("%016llx ",(*events)[i] );
   #endif

   /* restore affinty */
   sched_setaffinity(0, sizeof(cpu_set_t),&old_mask);
   #ifdef DEBUG 
    printf(" OK\n");fflush(stdout);
   #endif

   _uncore_unlock();
   return PAPI_OK;
}

/** free resources */
int UNCORE_shutdown(hwd_context_t *ctx)
{
   int i;

   #ifdef DEBUG 
    printf("UNCORE_shutdown...");fflush(stdout);
   #endif

   _uncore_lock();

   if (uncore_native_table!=NULL) free(uncore_native_table);
   if (package_data!=NULL)
   {
     for (i=0;i<num_packages;i++)
     {
       if (package_data[i].pc!=NULL) free(package_data[i].pc);
       if (package_data[i].pd!=NULL) free(package_data[i].pd);
     }
     free(package_data);
   }
   if (package_id!=NULL) free(package_id);
   if (contexts!=NULL) free(contexts);
   if (native_events!=NULL) free(native_events);

   _uncore_unlock();

   #ifdef DEBUG 
    printf(" OK\n");fflush(stdout);
   #endif
   return PAPI_OK;
}

/* This function sets various options in the substrate
 * The valid codes being passed in are PAPI_SET_DEFDOM,
 * PAPI_SET_DOMAIN, PAPI_SETDEFGRN, PAPI_SET_GRANUL * and PAPI_SET_INHERIT
 */
int UNCORE_ctl(hwd_context_t *ctx, int code, _papi_int_option_t *option)
{
   #ifdef DEBUG 
    printf("UNCORE_ctl...");fflush(stdout);
   #endif
   /*TODO implement? */
   #ifdef DEBUG 
    printf(" OK\n");fflush(stdout);
   #endif
   return PAPI_OK;
}

/*
 * setup counters
 * called when events are added to an eventset (native includes all events not only the added one)
 */
int UNCORE_update_control_state(hwd_control_state_t *ptr, NativeInfo_t *native, int count, hwd_context_t *ctx){
   int i,j,k,tmp;
   int index,ev_mask;                               /* index is the pfm event associated with the eventcode, ev_mask is the unit_mask */
   int selected;                                    /* used to check if an event was selected multiple times */
   cpu_set_t mask,old_mask;                         /* change and restore affinity */
   unsigned int num_masks=0,unit_mask,tmp_mask;     /* extract unit masks from event code */
   char event_name[PAPI_MAX_STR_LEN];               /* event name */
   pfarg_pmc_t *old_pmc;
   pfarg_pmd_t *old_pmd;
   unsigned int old_pmd_count,old_pmc_count;
   
   _uncore_lock();

   #ifdef DEBUG 
    printf("UNCORE_update_control_state - context: %p - hw_state: %p ...",ctx,ptr);fflush(stdout);
   #endif

   /* find duplicated context for a thread in case of multiple threads sharing an eventset */
   if (((UNCORE_context_t*)ctx)->thread_id!=PAPI_thread_id()) ctx=get_context_by_thread_id(((UNCORE_context_t*)ctx),PAPI_thread_id());

   /* copy selected events into data structure of context */
   if (((UNCORE_context_t*)ctx)->native_event_count==NULL) ((UNCORE_context_t*)ctx)->native_event_count=(int*)malloc(sizeof(int));
   *(((UNCORE_context_t*)ctx)->native_event_count)=count;
   ((UNCORE_context_t*)ctx)->native_events=realloc(((UNCORE_context_t*)ctx)->native_events,(*(((UNCORE_context_t*)ctx)->native_event_count))*sizeof(NativeInfo_t));   
   memcpy(((UNCORE_context_t*)ctx)->native_events,native,(*(((UNCORE_context_t*)ctx)->native_event_count))*sizeof(NativeInfo_t));

   /* construct superset of event sets of all contexts */
   tmp=0;
   for (i=0;i<num_contexts;i++) tmp+=*(contexts[i]->native_event_count);
   native_events=realloc(native_events,tmp*sizeof(NativeInfo_t));
   for (i=0;i<num_contexts;i++)
   {
     for (j=0;j<*(contexts[i]->native_event_count);j++)
     {
       index = (contexts[i]->native_events[j].ni_event & PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)>>8;
       ev_mask = (contexts[i]->native_events[j].ni_event & PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)& 0xff;

       pfm_get_num_event_masks(uncore_native_table[index].event, &num_masks);
       tmp_mask=0;
       for (k=0;k<num_masks;k++)
       {
         pfm_get_event_mask_code(uncore_native_table[index].event, k, &unit_mask);
         tmp_mask|=unit_mask;
       }
       /* remove invalid bits in mask */
       ev_mask&=tmp_mask;

       selected=0;
       for (k=0;k<native_event_count;k++) if (((native_events[k].ni_event& PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)>>8) ==index)
       { 
         selected=1;
         if (((native_events[k].ni_event& PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)& 0xff) !=ev_mask)
         {
           fprintf(stderr,"Error: cannot measure same event with different umasks simultainiously %02x -%02x\n",((native_events[k].ni_event& PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)& 0xff),ev_mask);
           _uncore_unlock();
           return PAPI_ECNFLCT;
         }
       }
       if (selected==0)
       {
         native_event_count++;
         native_events=realloc(native_events,native_event_count*sizeof(NativeInfo_t));
         memcpy(&(native_events[native_event_count-1]),&(contexts[i]->native_events[j]),sizeof(NativeInfo_t));
       }
     }
   }

   if (native_event_count>max_counters)
   {
     #ifdef DEBUG
      printf("Error: too many events selected\n");
     #endif
     _uncore_unlock();
     return PAPI_ECNFLCT;
   }

   #ifdef DEBUG
    printf("found %i events\n",native_event_count);
   #endif

   /* program event information into perfmon2 input parameters */
   for (j=0;j<native_event_count;j++)
   {
     index = (native_events[j].ni_event & PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)>>8;
     ev_mask = (native_events[j].ni_event & PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)& 0xff;

     pfm_get_num_event_masks(uncore_native_table[index].event, &num_masks);
     tmp_mask=0;
     for (i=0;i<num_masks;i++)
     {
       pfm_get_event_mask_code(uncore_native_table[index].event, i, &unit_mask);
       tmp_mask|=unit_mask;
     }
     /* remove invalid bits in mask */
     ev_mask&=tmp_mask;

     native_events[j].ni_position=j;
     for(i=0;i<count;i++) if (native_events[j].ni_event==native[i].ni_event) native[i].ni_position=j;

     for (k=0;k<num_packages;k++)
     {
       //TODO replace with pfm_find_event_bycode + set unit mask
       UNCORE_ntv_code_to_name((index<<8)|ev_mask, event_name, PAPI_MAX_STR_LEN-1);
       if (pfm_find_full_event(&(event_name[8]), &(package_data[k].inp.pfp_events[j]) ) != PFMLIB_SUCCESS)
       {
          printf("Error: cannot find event %i with unit_mask: %04x\n",uncore_native_table[index].event,ev_mask);
          printf("Error: pfm_find_full_event failed for: %s\n",&(event_name[8]));
          _uncore_unlock();
          return PAPI_EINVAL;
       }
       #ifdef DEBUG
       else
       {
         printf("added event \"%s\" on package %i\n",event_name,k);
       }
       #endif
       package_data[k].inp.pfp_events[j].plm=PFM_PLM3|PFM_PLM0;
     }
   }

   /* program control and data registers */
   for (k=0;k<num_packages;k++)
   {
     package_data[k].inp.pfp_event_count = native_event_count;
     if (pfm_dispatch_events(&(package_data[k].inp), package_data[k].mod_inp, &(package_data[k].outp), package_data[k].mod_outp) != PFMLIB_SUCCESS) {
        perror("Error: cannot dispatch events");
        _uncore_unlock();
        return PAPI_EBUG;
     }

     old_pmc=package_data[k].pc;
     old_pmd=package_data[k].pd;
     old_pmc_count=package_data[k].outp.pfp_pmc_count;
     old_pmd_count=package_data[k].outp.pfp_pmd_count;

     package_data[k].pc=calloc(package_data[k].outp.pfp_pmc_count,sizeof(pfarg_pmc_t));
     package_data[k].pd=calloc(package_data[k].outp.pfp_pmd_count,sizeof(pfarg_pmd_t));

     for(i=0; i < package_data[k].outp.pfp_pmc_count; i++) {
        package_data[k].pc[i].reg_num   = package_data[k].outp.pfp_pmcs[i].reg_num;
        package_data[k].pc[i].reg_value = package_data[k].outp.pfp_pmcs[i].reg_value;
     }
     /* another context is already running, stop it and restart counting with changed eventset */
     if (pfm_write_pmcs(package_data[k].fd, package_data[k].pc, package_data[k].outp.pfp_pmc_count) != PFMLIB_SUCCESS) {
        if (package_data[k].loaded_core!=-1){
          sched_getaffinity(0, sizeof(cpu_set_t),&old_mask);

          CPU_ZERO(&mask);
          CPU_SET(package_data[k].loaded_core,&mask);
          sched_setaffinity(0, sizeof(cpu_set_t),&mask);

          if (pfm_stop(package_data[k].fd)!=0) perror("UNCORE_stop pfm_stop");   
          if (pfm_unload_context(package_data[k].fd)!=0) perror("UNCORE_stop pfm_unload_context");

          if (pfm_write_pmcs(package_data[k].fd, package_data[k].pc, package_data[k].outp.pfp_pmc_count) != PFMLIB_SUCCESS) {
           perror("Error: pfm_write_pmcs failed");
           _uncore_unlock();
           return PAPI_EBUG;
          }
          else
          {
            if (pfm_load_context(package_data[k].fd, &(package_data[k].load_arg)) != PFMLIB_SUCCESS) {
              printf("pfm_load_context failed: fd=%i\n",package_data[k].fd);
              perror("");
              _uncore_unlock();
              return PAPI_EBUG;
            }
            else pfm_start(package_data[k].fd, NULL);
            sched_setaffinity(0, sizeof(cpu_set_t),&old_mask);
          }
        }
     }
     for(i=0; i < package_data[k].outp.pfp_pmd_count; i++) {
         package_data[k].pd[i].reg_num   = package_data[k].outp.pfp_pmds[i].reg_num;
         package_data[k].pd[i].reg_value = 0;
     }
     /* another context is already running, stop it and restart counting with changed eventset */
     if (pfm_write_pmds(package_data[k].fd, package_data[k].pd, package_data[k].outp.pfp_pmd_count) != PFMLIB_SUCCESS) {
        if (package_data[k].loaded_core!=-1){
          sched_getaffinity(0, sizeof(cpu_set_t),&old_mask);

          CPU_ZERO(&mask);
          CPU_SET(package_data[k].loaded_core,&mask);
          sched_setaffinity(0, sizeof(cpu_set_t),&mask);

          if (pfm_stop(package_data[k].fd)!=0) perror("UNCORE_stop pfm_stop");
          if (pfm_unload_context(package_data[k].fd)!=0) perror("UNCORE_stop pfm_unload_context");
          /* restore last_values in package_data[k].pd[i].reg_value */
          if ((package_data[k].outp.pfp_pmd_count==package_data[k].outp.pfp_pmc_count)&&(old_pmd_count==old_pmc_count))
            for(i=0; i < package_data[k].outp.pfp_pmd_count; i++)
             for (j=0; j < old_pmc_count; j++)
               if (old_pmc[j].reg_value == package_data[k].pc[i].reg_value) package_data[k].pd[i].reg_value = old_pmd[j].reg_value;


          if (pfm_write_pmds(package_data[k].fd, package_data[k].pd, package_data[k].outp.pfp_pmd_count) != PFMLIB_SUCCESS) {
            printf("Error: pfm_write_pmds failed\n");
            perror("");
            _uncore_unlock();
            return PAPI_EBUG;
          }
          else
          {
            if (pfm_load_context(package_data[k].fd, &(package_data[k].load_arg)) != PFMLIB_SUCCESS) {
              printf("pfm_load_context failed: fd=%i\n",package_data[k].fd);
              perror("");
              _uncore_unlock();
              return PAPI_EBUG;
            }
            else pfm_start(package_data[k].fd, NULL);
            sched_setaffinity(0, sizeof(cpu_set_t),&old_mask);
          }
        }
     }

     free(old_pmc);
     free(old_pmd);
   }
   _uncore_unlock();
   #ifdef DEBUG 
    printf("OK\n");fflush(stdout);
   #endif
   return PAPI_OK;
}

/*
 * This function has to set the bits needed to count different domains
 * In particular: PAPI_DOM_USER, PAPI_DOM_KERNEL PAPI_DOM_OTHER
 * By default return PAPI_EINVAL if none of those are specified
 * and PAPI_OK with success
 * PAPI_DOM_USER is only user context is counted
 * PAPI_DOM_KERNEL is only the Kernel/OS context is counted
 * PAPI_DOM_OTHER  is Exception/transient mode (like user TLB misses)
 * PAPI_DOM_ALL   is all of the domains
 */
int UNCORE_set_domain(hwd_control_state_t *cntrl, int domain) 
{
  int found = 0;

  /*if ( PAPI_DOM_USER & domain ){
        found = 1;
  }
  if ( PAPI_DOM_KERNEL & domain ){
        found = 1;
  }
  if ( PAPI_DOM_OTHER & domain ){
        found = 1;
  }*/
  if ( PAPI_DOM_ALL == domain ){
        found = 1;
  }
  if ( !found )
        return PAPI_EINVAL;
   return PAPI_OK;
}

/*
 * (PAPI_reset() was called), simply use current counter values as start values of that context
 */
int UNCORE_reset(hwd_context_t *ctx, hwd_control_state_t *ctrl)
{
   int i;

   #ifdef DEBUG 
    printf("UNCORE_reset...");fflush(stdout);
   #endif

   _uncore_lock();

   if (((UNCORE_context_t*)ctx)->thread_id!=PAPI_thread_id()) ctx=get_context_by_thread_id(((UNCORE_context_t*)ctx),PAPI_thread_id());

   #ifdef DEBUG 
    assert(package_data[((UNCORE_context_t*)ctx)->package_id].monitoring!=0);
   #endif
   if (update_counters(package_data[((UNCORE_context_t*)ctx)->package_id].monitoring)!=0) {_uncore_unlock();return PAPI_EBUG;}
   for (i=0;i<UNCORE_MAX_COUNTERS;i++)((UNCORE_context_t*)ctx)->start_values[i]=package_data[((UNCORE_context_t*)ctx)->package_id].last_results[i];

   _uncore_unlock();

   #ifdef DEBUG 
    printf(" OK\n");fflush(stdout);
   #endif
   return PAPI_OK;
}

/************************** Native Event functions ***************************/

/** iterate through native event table until invalid event code is reached */
int UNCORE_ntv_enum_events(unsigned int *EventCode, int modifier)
{
   int cidx = PAPI_COMPONENT_INDEX(*EventCode);
   int index,mask,i;
   unsigned int tmp,num_masks,unit_mask;

   switch(modifier) {
      case PAPI_ENUM_FIRST:
         *EventCode = PAPI_NATIVE_MASK|PAPI_COMPONENT_MASK(cidx);
         return (PAPI_OK);
         break;
      case PAPI_ENUM_EVENTS: {
         index = (*EventCode & PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)>>8;

         if (index < (NUM_EVENTS>>8)) {
           
           /* // instead of do{}while; loop if ENUM_UMASKS would be used by _papi_hwi_native_name_to_code
            * *EventCode = (*EventCode - ((*EventCode)&0xff)) + 0x100; 
            * index = (*EventCode & PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)>>8;
            */

           do
           {             
             *EventCode = *EventCode  + 1;

             index = (*EventCode & PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)>>8;
             mask = (*EventCode & PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)& 0xff;
             pfm_get_num_event_masks(uncore_native_table[index].event, &num_masks);
             tmp=0;
             for (i=0;i<num_masks;i++)
             {
                pfm_get_event_mask_code(uncore_native_table[index].event, i, &unit_mask);
                tmp|=unit_mask;
             }
           }
           while((((mask&tmp)!=mask))&&(index<(NUM_EVENTS>>8))); /* skip over invalid unit masks */

           if (index < (NUM_EVENTS>>8)) return (PAPI_OK); else return (PAPI_ENOEVNT);
         } else {
            return (PAPI_ENOEVNT);
         }
         break;
      }
      /* unfortunately not used by _papi_hwi_native_name_to_code */
      /*case PAPI_NTV_ENUM_UMASKS:{
        do
        {
          printf("enum umasks\n");
          *EventCode = *EventCode + 1;
          index = (*EventCode & PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)>>8;
          mask = (*EventCode & PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)& 0xff;
          pfm_get_num_event_masks(uncore_native_table[index].event, &num_masks);
          
          for (i=0;i<num_masks;i++)
          {
             pfm_get_event_mask_code(uncore_native_table[index].event, i, &unit_mask);
             if (((*EventCode)&0xff)==unit_mask) return (PAPI_OK);
          }
        }
        while (((*EventCode)&0xff)<0xff);
        return (PAPI_EINVAL);
      }*/
      default: 
         return (PAPI_EINVAL);
   }
   return (PAPI_EINVAL);
}

/** translate event code into human readable name */
int UNCORE_ntv_code_to_name(unsigned int EventCode, char *name, int len)
{
   unsigned int num_masks=0;
   unsigned int unit_mask,i,tmp;
   int index = (EventCode & PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)>>8;
   int mask = (EventCode & PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)& 0xff;
   char event_name[PAPI_MAX_STR_LEN];
   char event_mask_name[PAPI_MAX_STR_LEN];

   #ifdef DEBUG 
    printf("UNCORE_code_to_name...");fflush(stdout);
   #endif

   pfm_get_num_event_masks(uncore_native_table[index].event, &num_masks);
   tmp=0;
   for (i=0;i<num_masks;i++)
   {
     pfm_get_event_mask_code(uncore_native_table[index].event, i, &unit_mask);
     tmp|=unit_mask;
   }
   /* remove invalid bits in mask */
   mask&=tmp;
   
   //TODO get rid of this (len is reported to be 1024 but writing more than 128 seems to cause buffer overflows);
   if (len>128) len=128;

   pfm_get_event_name(uncore_native_table[index].event, event_name, PAPI_MAX_STR_LEN-1);

   if ((index<(NUM_EVENTS>>8))&&(!mask))
   {
     snprintf(name,len,"SYSWIDE_%s",event_name);
     return(PAPI_OK);
   }

   if ((index<(NUM_EVENTS>>8))&&(mask))
   { 
     snprintf(name,len,"SYSWIDE_%s",event_name);
     for (i=0;i<num_masks;i++)
     {
       pfm_get_event_mask_code(uncore_native_table[index].event, i, &unit_mask);
       pfm_get_event_mask_name(uncore_native_table[index].event, i, event_mask_name, PAPI_MAX_STR_LEN-1);
       if ((unit_mask<=mask)&&(unit_mask&mask))
       {
          if (len>strlen(name)+2+strlen(event_mask_name)) 
          {
            strncat(name,":",len-strlen(name));
            strncat(name,event_mask_name,len-strlen(name));
          }
       }
       if (unit_mask==mask) {snprintf(name,len,"SYSWIDE_%s:%s",event_name,event_mask_name);break;} 
     }
     return(PAPI_OK);
   }
   
   return(PAPI_EINVAL);
}

/** get description of the event code (including unit mask)*/
int UNCORE_ntv_code_to_descr(unsigned int EventCode, char *name, int len)
{
   unsigned int num_masks=0;
   unsigned int unit_mask,i,tmp;
   int index = (EventCode & PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)>>8;
   int mask = (EventCode & PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)& 0xff;
   char* description=NULL;

   #ifdef DEBUG 
    printf("UNCORE_code_to_descr...");fflush(stdout);
   #endif

   pfm_get_num_event_masks(uncore_native_table[index].event, &num_masks);
   tmp=0;
   for (i=0;i<num_masks;i++)
   {
     pfm_get_event_mask_code(uncore_native_table[index].event, i, &unit_mask);
     tmp|=unit_mask;
   }
   /* remove invalid bits in mask */
   mask&=tmp;

   //TODO get rid of this (len is reported to be 1024 but writing more than 128 seems to cause buffer overflows);
   if (len>128) len=128;

   /* event description */
   pfm_get_event_description(uncore_native_table[index].event, &description);
   strncpy(name, description, len);
   if (description!=NULL) free(description);

   if ((index<(NUM_EVENTS>>8))&&(!mask))
   {
     return(PAPI_OK);
   }

   if ((index<(NUM_EVENTS>>8))&&(mask==tmp))
   {
     for (i=0;i<num_masks;i++)
     {
       pfm_get_event_mask_code(uncore_native_table[index].event, i, &unit_mask);
       if (unit_mask==tmp)
       {
          pfm_get_event_mask_description(uncore_native_table[index].event, i, &description);
          //strncat(name,":",len-strlen(name));strncat(name,description,len-strlen(name));
          if (description!=NULL) free(description);
          return(PAPI_OK);
       }
     }
   }

   if ((index<(NUM_EVENTS>>8))&&(mask))
   { 
     for (i=0;i<num_masks;i++)
     {
       pfm_get_event_mask_code(uncore_native_table[index].event, i, &unit_mask);
       if ((unit_mask&mask)&&(unit_mask!=tmp))
       {
         pfm_get_event_mask_description(uncore_native_table[index].event, i, &description);
         //strncat(name,":",len-strlen(name));strncat(name,description,len-strlen(name));
         if (description!=NULL) free(description);
       }
     }
     return(PAPI_OK);
   }
     
   return(PAPI_EINVAL);
        
}

/** 
  * return copy of hwd_register_t struct associated with eventcode 
  */
int UNCORE_ntv_code_to_bits(unsigned int EventCode, hwd_register_t * bits)
{
   int index = (EventCode & PAPI_NATIVE_AND_MASK & PAPI_COMPONENT_AND_MASK)>>8;

   if (index<(NUM_EVENTS>>8))
   {
     memcpy( ( UNCORE_register_t *) bits,
         &(uncore_native_table[index].resources),
         sizeof(UNCORE_register_t) );
     return (PAPI_OK);   
   }
   
   else return(PAPI_EINVAL);
        
}

/*
 * function vector used by PAPI-C to access component
 * copy and paste from ACPI example component
 * TODO ??? implement missing functions:
 *   - UNCORE_ctl (exists but empty)
 *   - ntv_bits_to_info
 *   - bpt_map_*
 * TODO cmp_info initialization - attach/attach_must_ptrace ???
 */
papi_vector_t _uncore_vector = {
   .cmp_info = {
   /* default component information (unspecified values are initialized to 0)*/
   .name =                    "$Id: UNCORE component 01/20/2010$",
   .version =                 "$Revision: 0.13$",
   .num_mpx_cntrs =           PAPI_MPX_DEF_DEG,
   .num_cntrs =               UNCORE_MAX_COUNTERS,
   .default_domain =          PAPI_DOM_ALL,
   .available_domains =       PAPI_DOM_ALL,
   .default_granularity =     PAPI_GRN_THR,
   .available_granularities=  PAPI_GRN_THR,
   .hardware_intr_sig =       PAPI_INT_SIGNAL,

   /* component specific cmp_info initializations */
   .fast_real_timer =         0,
   .fast_virtual_timer =      0,
   .attach =                  0,
   .attach_must_ptrace =      0,
   },

   /* sizes of framework-opaque component-private structures */
   .size = {
   .context =                 sizeof(UNCORE_context_t),
   .control_state =           sizeof(UNCORE_control_state_t),
   .reg_value =               sizeof(UNCORE_register_t),
   .reg_alloc =               sizeof(UNCORE_reg_alloc_t),
   },
   /* function pointers in this component */
   .init =                    UNCORE_init,
   .init_substrate =          UNCORE_init_substrate,
   .init_control_state =      UNCORE_init_control_state,
   .start =                   UNCORE_start,
   .stop =                    UNCORE_stop,
   .read =                    UNCORE_read,
   .shutdown =                UNCORE_shutdown,
   .ctl =                     UNCORE_ctl,
   .bpt_map_set =             NULL,
   .bpt_map_avail =           NULL,
   .bpt_map_exclusive =       NULL,
   .bpt_map_shared =          NULL,
   .bpt_map_preempt =         NULL,
   .bpt_map_update =          NULL,

   .update_control_state =    UNCORE_update_control_state,
   .set_domain =              UNCORE_set_domain,
   .reset =                   UNCORE_reset,

   .ntv_enum_events =         UNCORE_ntv_enum_events,
   .ntv_code_to_name =        UNCORE_ntv_code_to_name,
   .ntv_code_to_descr =       UNCORE_ntv_code_to_descr,
   .ntv_code_to_bits =        UNCORE_ntv_code_to_bits,
   .ntv_bits_to_info =        NULL,
};

